#include <moveit/move_group_interface/move_group_interface.h>
#include <moveit/planning_scene_interface/planning_scene_interface.h>
#include <moveit_msgs/DisplayRobotState.h>
#include <moveit_msgs/DisplayTrajectory.h>
#include <moveit_msgs/AttachedCollisionObject.h>
#include <moveit_msgs/CollisionObject.h>
#include <moveit_visual_tools/moveit_visual_tools.h>

#include <geometry_msgs/Pose.h>
#include <std_msgs/String.h>

#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

#include <remote_rosbag_record/call.hpp>
#include <boost/regex.hpp>

#include <tf2_ros/buffer.h>
#include <tf2_ros/transform_listener.h>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>

#include <boost/algorithm/string.hpp>
#include <iostream>
#include <fstream>
#include <string>

#include <typeinfo>

#define foreach BOOST_FOREACH



class CSVReader{

  std::string fileName;
  std::string delimeter;

public:
  CSVReader(std::string filename, std::string delm = ",") :
	  fileName(filename), delimeter(delm)
  { }

  //Function to fetch data
  std::vector<std::vector<std::string> > getData();
};


std::vector<std::vector<std::string> > CSVReader::getData(){
	
  std::ifstream file(fileName);

  std::vector<std::vector<std::string> >dataList;
  std::string line = "";

  //iterate through lines and split content
  while (getline(file, line)){
           
	std::vector<std::string> vec;
	boost::algorithm::split(vec, line, boost::is_any_of(delimeter));
  	dataList.push_back(vec);
  
  }


  file.close();

  return dataList;

}

class robotJoints{


public:	std::vector<double> joints;

	robotJoints() {
		joints = {0,0,0,0,0,0};
	}
	robotJoints(std::vector<double> input_joints){
		joints = input_joints;
       	}

	//functions
	
};




int main(int argc, char** argv)
{
  ros::init(argc, argv, "measurements");
  ros::NodeHandle node_handle;

  // ROS spinning must be running for the MoveGroupInterface to get information
  // about the robot's state. One way to do this is to start an AsyncSpinner
  // beforehand.
  ros::AsyncSpinner spinner(1);
  spinner.start();


  // Setup
  // ^^^^^
  //
  // MoveIt operates on sets of joints called "planning groups" and stores them in an object called
  // the `JointModelGroup`. Throughout MoveIt the terms "planning group" and "joint model group"
  // are used interchangably.
  static const std::string PLANNING_GROUP = "manipulator";
  
  // The :planning_interface:`MoveGroupInterface` class can be easily
  // setup using just the name of the planning group you would like to control and plan for.
  moveit::planning_interface::MoveGroupInterface move_group_interface(PLANNING_GROUP);
  move_group_interface.setPlanningTime(15.0);

  // We will use the :planning_interface:`PlanningSceneInterface`
  // class to add and remove collision objects in our "virtual world" scene
  moveit::planning_interface::PlanningSceneInterface planning_scene_interface;

  // Raw pointers are frequently used to refer to the planning group for improved performance. 
  const moveit::core::JointModelGroup* joint_model_group =
      move_group_interface.getCurrentState()->getJointModelGroup(PLANNING_GROUP);

  // Visualization
  // ^^^^^^^^^^^^^
  //
  // The package MoveItVisualTools provides many capabilities for visualizing objects, robots,
  // and trajectories in RViz as well as debugging tools such as step-by-step introspection of a script.
  namespace rvt = rviz_visual_tools;
  moveit_visual_tools::MoveItVisualTools visual_tools("base_link");
  visual_tools.deleteAllMarkers();

  // Remote control is an introspection tool that allows users to step through a high level script
  // via buttons and keyboard shortcuts in RViz
  visual_tools.loadRemoteControl();

  // RViz provides many types of markers, in this demo we will use text, cylinders, and spheres
  Eigen::Isometry3d text_pose = Eigen::Isometry3d::Identity();
  text_pose.translation().z() = 1.0;

  // Batch publishing is used to reduce the number of messages being sent to RViz for large visualizations
  visual_tools.trigger();

  // Getting Basic Information
  // ^^^^^^^^^^^^^^^^^^^^^^^^^
  //
  // We can print the name of the reference frame for this robot.
  ROS_INFO_NAMED("tutorial", "Planning frame: %s", move_group_interface.getPlanningFrame().c_str());

  // We can also print the name of the end-effector link for this group.
  ROS_INFO_NAMED("tutorial", "End effector link: %s", move_group_interface.getEndEffectorLink().c_str());

  // We can get a list of all the groups in the robot:
  ROS_INFO_NAMED("tutorial", "Available Planning Groups:");
  std::copy(move_group_interface.getJointModelGroupNames().begin(),
            move_group_interface.getJointModelGroupNames().end(), std::ostream_iterator<std::string>(std::cout, ", "));



  // Start 
  // ^^^^^
  //
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to start the demo, insert the scene");

  // Note that we are just planning, not asking move_group_interface
  // to actually move the robot.
  moveit::planning_interface::MoveGroupInterface::Plan my_plan;

  

  // Adding objects to the environment
  // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  //

  // Define a collision object ROS message for the robot to avoid.
  
  moveit_msgs::CollisionObject co0;
  moveit_msgs::CollisionObject co1;
  moveit_msgs::CollisionObject co2;

  co0.header.frame_id = move_group_interface.getPlanningFrame();
  co1.header.frame_id = move_group_interface.getPlanningFrame();
  co2.header.frame_id = move_group_interface.getPlanningFrame();   
 
  // The id of the object is used to identify it
  co0.id = "box0";
  co1.id = "box1";  
  co2.id = "box2";

  // Define box0 to add to the world.
  shape_msgs::SolidPrimitive primitive;
  primitive.type = primitive.BOX;
  primitive.dimensions.resize(3);
  primitive.dimensions[primitive.BOX_X] = 1.76;
  primitive.dimensions[primitive.BOX_Y] = 0.7;
  primitive.dimensions[primitive.BOX_Z] = 0.03;
  // Define a pose for box0 (specified relative to frame_id)
  geometry_msgs::Pose box_pose;
  box_pose.orientation.w = 1.0;
  box_pose.position.x = -0.66;
  box_pose.position.y = 0.21;
  box_pose.position.z = -0.01;                                   
  co0.primitives.push_back(primitive);
  co0.primitive_poses.push_back(box_pose);
  co0.operation = co0.ADD;
                           

  // Define box1 to add to the world.
  primitive.dimensions[primitive.BOX_X] = 0.88;
  primitive.dimensions[primitive.BOX_Y] = 0.88;
  primitive.dimensions[primitive.BOX_Z] = 0.03;
  // Define a pose for box1 (specified relative to frame_id)
  box_pose.position.x = -1.1;
  box_pose.position.y = 1;
  box_pose.position.z = -0.01;                                         
  co1.primitives.push_back(primitive);
  co1.primitive_poses.push_back(box_pose);
  co1.operation = co1.ADD;


  // Define box2 to add to the world.
  primitive.dimensions[primitive.BOX_X] = 0.15;
  primitive.dimensions[primitive.BOX_Y] = 0.07;
  primitive.dimensions[primitive.BOX_Z] = 0.3;
  // Define a pose for box2 (specified relative to frame_id)
  box_pose.position.x = 0.0;
  box_pose.position.y = -0.18;
  box_pose.position.z = -0.15;
  co2.primitives.push_back(primitive);
  co2.primitive_poses.push_back(box_pose);
  co2.operation = co2.ADD;
  
  //all the collision objects
  std::vector<moveit_msgs::CollisionObject> collision_objects;
  collision_objects.push_back(co0);
  collision_objects.push_back(co1);
  collision_objects.push_back(co2);  


  // Now, let's add the collision object into the world
  // (using a vector that could contain additional objects)
  ROS_INFO_NAMED("tutorial", "Add objects into the world");
  planning_scene_interface.addCollisionObjects(collision_objects);
  



  // Read csv file and store joint states
  // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  // 
  // *no header in csv, values to be double, in radians
  //
  
  visual_tools.trigger();
  visual_tools.prompt("Press 'next' to in the RvizVisualToolsGui to read from the csv");

  std::string traj_file = "testvals_smaller";
  CSVReader reader("/home/ioana/catkin_ws/src/robot_identification/csv_files/"+traj_file+".csv");
    
  std::vector<std::vector<std::string> > csvData = reader.getData();

  //create vector of robotjoints to store all joint states
  std::vector<robotJoints> allJoints;
  std::vector<double> auxVect(6);
    
  for (std::vector<std::string> vec:csvData){

        //changing strings from vec to double in auxVect	  
	std::transform(vec.begin(), vec.end(), auxVect.begin(), [](const std::string& val)
        { return std::stod(val); }    
	);

	allJoints.push_back(robotJoints(auxVect));
		
  }

  // check joints
  foreach( robotJoints r, allJoints){
        
        std::cout << std::endl << "New joint state: " << std::endl;
	for (int i=0; i<6; i++ ){
		std::cout << r.joints[i] << ", ";
	}
  } 


   


  // Planning with the acquired values from the csv
  // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 
  visual_tools.trigger();
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to plan going to a joint state ");

  
  //Current robot state
  moveit::core::RobotStatePtr current_state = move_group_interface.getCurrentState();
  
  // Next get the current set of joint values for the group.
  std::vector<double> joint_group_positions;
  current_state->copyJointGroupPositions(joint_model_group, joint_group_positions);

  robot_state::RobotState current_start_state(*move_group_interface.getCurrentState());
  move_group_interface.setStartState(current_start_state);
  robot_state::RobotState new_start_state(current_start_state);
    
  int i = 0;
  foreach (robotJoints r, allJoints){
    //for each line read from csv, set those values as the joint targets and plan

    joint_group_positions = r.joints; 
    move_group_interface.setJointValueTarget(joint_group_positions);

    // Set your preferred maximum defaults in the joint_limits.yaml file of your robot's moveit_config. This scaling factor should be the same as with the teaching pad percentage at runtime 
    move_group_interface.setMaxVelocityScalingFactor(1.0);
    move_group_interface.setMaxAccelerationScalingFactor(1.0);

    bool success = (move_group_interface.plan(my_plan) == moveit::planning_interface::MoveItErrorCode::SUCCESS);
    ROS_INFO_NAMED("tutorial", "Planning to joint space goal %s", success ? "" : "FAILED");

    // Visualize the plan in RViz
    visual_tools.deleteAllMarkers();
    visual_tools.publishText(text_pose, "Joint Space Goal", rvt::WHITE, rvt::XLARGE);
    visual_tools.publishTrajectoryLine(my_plan.trajectory_, joint_model_group);
 
    new_start_state.setJointGroupPositions(joint_model_group, r.joints); 
    move_group_interface.setStartState(new_start_state);

 


  // Recording and Execution  -  to be done for each line, since my_plan is updating each iteration
  // ^^^^^^^^^^^^^^^^^^^^^^^
  
  // path of the folder the bag files will be saved in
  std::string path = "/home/ioana/catkin_ws/measurements_bag_files/";
  std::string filename = traj_file + std::to_string(i);
  // changing value in parameter server for the logging of the bag files
  node_handle.setParam("/robot_record/prefix",path+filename);
  i++;



  visual_tools.trigger();             
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window only if you want to go with the execution");

  remote_rosbag_record::call(boost::regex("/start"));
  ros::Duration(2).sleep();
  move_group_interface.execute(my_plan);
  ros::Duration(2).sleep();
  remote_rosbag_record::call(boost::regex("/stop"));

  ros::Duration(3).sleep();
  
  }



  // Remove the objects from the world.
  // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 
  visual_tools.trigger();                                             
  visual_tools.prompt("Press 'next' to continue, removing objects");  
 
  ROS_INFO_NAMED("tutorial", "Remove the objects from the world");
  std::vector<std::string> object_ids;
  object_ids.push_back(co0.id);
  object_ids.push_back(co1.id);
  object_ids.push_back(co2.id);
  planning_scene_interface.removeCollisionObjects(object_ids);

  visual_tools.trigger();
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to once the collision object disapears");

                                          
  
  // END


  ros::shutdown();
  return 0;
}
