#include <pluginlib/class_loader.h>
#include <ros/ros.h>
#include <ros/service_client.h>
#include <ros/service.h>

#include <boost/scoped_ptr.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <moveit_visual_tools/moveit_visual_tools.h>

#include <remote_rosbag_record/call.hpp>
#include <boost/regex.hpp>

#include <iostream>
#include <signal.h>
#include <string>
#include <robot_state_publisher/robot_state_publisher.h> 
#include <std_msgs/Float64MultiArray.h>
#include <controller_manager_msgs/SwitchController.h>
#include <trajectory_msgs/JointTrajectory.h>
#include <trajectory_msgs/JointTrajectoryPoint.h>
#include <std_msgs/Float64.h>
#include <std_srvs/Trigger.h>

#include <dirent.h>

#define foreach BOOST_FOREACH

using namespace std;



class CSVReader{

  std::string fileName;
  std::string delimeter;

public:
  CSVReader(std::string filename, std::string delm = ",") :
          fileName(filename), delimeter(delm)
  { }

  //Function to fetch data
  std::vector<std::vector<std::string> > getData();
};


std::vector<std::vector<std::string> > CSVReader::getData(){

  std::ifstream file(fileName);

  std::vector<std::vector<std::string> >dataList;
  std::string line = "";

  //iterate through lines and split content
  while (getline(file, line)){

        std::vector<std::string> vec;
        boost::algorithm::split(vec, line, boost::is_any_of(delimeter));
        dataList.push_back(vec);

  }


  file.close();

  return dataList;

}


// class containing, velocity, angle to be travelled (in radians), number of repetitions and pose to start from

class robotData{
public: std::vector<double> home;
	double velocity;
	double angle;
	double reps;
/*
        robotData() {
                home = {0,-1.57,0,-1.57,0,0};
		velocity = 0;
		angle = 0;
		reps = 1;
        }
*/
        robotData(std::vector<double> in_home, double in_vel, double in_angle, double in_reps){

                home = in_home;
		velocity = in_vel;
		angle = in_angle;
		reps = in_reps;
        }

};

void node_shutdown(int sig)
{
  ROS_INFO_STREAM("Shutting down cleanly");
  std_srvs::Trigger stop;
  ros::service::call("/ur_hardware_interface/dashboard/stop",stop);
  ROS_INFO_STREAM("Stopped the URCaps program on the dashboard");
  ros::Duration(1).sleep();
  ros::shutdown();
}

int main(int argc, char** argv)
{
  const std::string node_name = "velocities_alone";
  ros::init(argc, argv, node_name, ros::init_options::NoSigintHandler);
  ros::AsyncSpinner spinner(1);
  spinner.start();
  ros::NodeHandle node_handle("~");
  signal(SIGINT, node_shutdown);

  // publisher for the velocity commands
  ros::Publisher joint_pub = node_handle.advertise<std_msgs::Float64MultiArray>("/joint_group_vel_controller/command",1); 
  std_msgs::Float64MultiArray command_velocities;

  // publisher for the home positin commands
  ros::Publisher home_pub = node_handle.advertise<trajectory_msgs::JointTrajectory>("/scaled_pos_joint_traj_controller/command",1);
  trajectory_msgs::JointTrajectory home_msg;
  trajectory_msgs::JointTrajectoryPoint home_point;




  // Find the directory containing the csv files to iterate through
  // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

  std::vector<std::string> files;
  struct dirent *entry;
  DIR *dir = opendir("/home/ioana/catkin_ws/src/robot_identification/csv_files/csv_joints/");

  while ((entry = readdir(dir)) !=NULL){
        files.push_back(entry->d_name);
        if (files.back().find(".csv") == std::string::npos ){
            files.pop_back();}
  }
  closedir(dir);

  //for each csv file found in the dir
  foreach (std::string csv_str, files){

    std::cout<<"Found file: "<<csv_str<<std::endl;


    // Read csv file and store robot data
    // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  
    //path and file to be read
    std::string csv_path = "/home/ioana/catkin_ws/src/robot_identification/csv_files/csv_joints/";
    CSVReader reader(csv_path+csv_str);
    std::vector<std::vector<std::string>> csvData = reader.getData();

    //create vector of robData objects to store all csv data
    std::vector<robotData> all_robData;
    std::vector<double> auxVect(9);
    std::vector<double> home(6);
    double vel;
    double angle;
    double reps;

    for (std::vector<std::string> vec:csvData){
        //changing strings from vec to double in auxVect          
        std::transform(vec.begin(), vec.end(), auxVect.begin(), [](const std::string& val)
        { return std::stod(val); }
        );
	
	vel   = auxVect[6];
	angle = auxVect[7];
	reps  = auxVect[8];
	home  = {auxVect.begin(), auxVect.begin()+6};

        all_robData.push_back(robotData(home, vel, angle, reps));
    }  

    // check values
    foreach( robotData rd, all_robData){
        std::cout << std::endl << "New state: " << std::endl;
        std::cout << ", velocity: "<< rd.velocity << ",angle: "<< rd.angle <<", reps: "<< rd.reps << std::endl;
        
    }

    //Executing and recording
    //^^^^^^^^^^^^^^^^^^^^^^^
    ros::service::waitForService("/controller_manager/switch_controller");

    ros::ServiceClient play_program = node_handle.serviceClient<std_srvs::Trigger>("/ur_hardware_interface/dashboard/play");
    std_srvs::Trigger play;
    play_program.call(play);
    ROS_INFO_STREAM("Played the URCaps program on the dashboard");
    ros::Duration(1).sleep();

    controller_manager_msgs::SwitchController::Request req;
    controller_manager_msgs::SwitchController::Response res;
    bool success;
    req.strictness = 2;
    req.start_asap = false;
    req.timeout = 0.0;

    // the for loop iterating through the vector of velocities/angles    
    foreach (robotData rd, all_robData){

      for (int r=0; r<rd.reps; r++){	    

      //pose to start from	    
      home_point.positions = rd.home;
      home_point.time_from_start.sec = 5;
      home_msg.joint_names = {"shoulder_pan_joint","shoulder_lift_joint","elbow_joint","wrist_1_joint","wrist_2_joint","wrist_3_joint"};
      home_msg.points = {home_point};

      home_pub.publish(home_msg);
      ros::Duration(6).sleep();

      // Change controller to joint_group_vel_controller by service call
      ros::service::waitForService("/controller_manager/switch_controller");

      // scaled_pos... is default for the respective launch file
      req.start_controllers = {"joint_group_vel_controller"};
      req.stop_controllers  = {"scaled_pos_joint_traj_controller"};
  
      success = ros::service::call("/controller_manager/switch_controller",req,res);
      if(success)
        ROS_INFO_STREAM("Switched to joint_vel_group_controller"<<res.ok);
      else{
        ROS_ERROR_STREAM("Failed to switch controllers, is the external control running?");
        return 0; }

      //visual_tools.trigger();
      //visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to execute");

      double angle   = rd.angle;
      double vel     = rd.velocity;    
      double thetime = angle/vel;

      // path of the folder the bag files will be saved in
      std::string path = "/home/ioana/catkin_ws/measurements_bag_files/newramps/";
      std::string filename = "j" + std::string(1,csv_str[1]) + "_" + to_string(vel)+"_"+to_string(r);
      // changing value in parameter server for the logging of the bag files 
      node_handle.setParam("/robot_record/prefix",path+filename);
      int joint_nr = int(csv_str[1]-'0')-1;

      std::cout<<"Joint to move "<<joint_nr+1<<" for time "<< thetime<< "with velocity"<<vel<<std::endl;

      remote_rosbag_record::call(boost::regex("/start"));
      ros::Duration(2).sleep();

      //first 0 point
      command_velocities.data = {0,0,0,0,0,0};
      joint_pub.publish(command_velocities);
      //first !=0 point    
      command_velocities.data[joint_nr] = vel;
      joint_pub.publish(command_velocities);
      ros::Duration(thetime).sleep();
      //2nd !=0 point
      joint_pub.publish(command_velocities);
      // 2nd 0 point
      command_velocities.data = {0,0,0,0,0,0};
      joint_pub.publish(command_velocities);

      ros::Duration(2).sleep();
      remote_rosbag_record::call(boost::regex("/stop"));

      //visual_tools.trigger();
      //visual_tools.prompt("Move robot home, press 'next' in the RvizVisualToolsGui window to continue");

      ros::Duration(2).sleep();
    
      // Change controller to scaled_pos_joint_traj_controller to be able to move to next home position
      ros::service::waitForService("/controller_manager/switch_controller");
      req.stop_controllers = {"joint_group_vel_controller"};
      req.start_controllers  = {"scaled_pos_joint_traj_controller"};
      
      success = ros::service::call("/controller_manager/switch_controller",req,res);
      if(success)
        ROS_INFO_STREAM("Switched to scaled_pos_joint_traj_controller"<<res.ok);
   
      //visual_tools.trigger();
      //visual_tools.prompt("Next movement, press 'next' in the RvizVisualToolsGui window to continue with next iteration");

      }
    }
  }

  ros::ServiceClient stop_program = node_handle.serviceClient<std_srvs::Trigger>("/ur_hardware_interface/dashboard/stop");
  std_srvs::Trigger stop;
  stop_program.call(stop);
  ROS_INFO_STREAM("Stopped the URCaps program on the dashboard");

  //END
  return 0;
}
